#ifndef WINDOWSTATS_H
#define WINDOWSTATS_H

#include <Sequence/FST.hpp>
#include <Sequence/PolyTableSlice.hpp>


double calculateGmin(Sequence::SimData & wpop1, Sequence::SimData & wpop2, bool verbose = false)
{
  unsigned nsegs = wpop1.segsites();
  if (nsegs != wpop2.segsites()){ std::cout<<"Warning, number of segregating sites do not match between pop1 and pop2!"<<std::endl; return 0.0; }
  std::vector<std::string> hap1 = wpop1.GetData();
  std::vector<std::string> hap2 = wpop2.GetData();

  unsigned n1 = hap1.size();
  unsigned n2 = hap2.size();
  std::vector<unsigned> diffs(n1*n2,0);
  
  unsigned comp = 0;
  for (auto i = hap1.cbegin(); i != hap1.cend(); ++i)
  {
    for (auto j = hap2.cbegin(); j != hap2.cend(); ++j)
    {
      for (unsigned seg = 0; seg < nsegs; ++seg) if ( (*i)[seg] != (*j)[seg] ) ++diffs[comp];
      ++comp;
    }
  }

  if (comp != n1*n2){ std::cerr<<"Warning, number of comparisons does not match ("<<comp<<" vs. "<<n1*n2<<")!"<<std::endl; return 0.0; }
  double dxy = std::accumulate(diffs.cbegin(), diffs.cend(), 0) / (double)comp;
  auto minitr = std::min_element(diffs.cbegin(), diffs.cend());
  if (verbose) std::cout<<"Dxy: "<<dxy<<", min dxy: "<<*minitr<<", Gmin: "<<(*minitr)/dxy<<std::endl;
  return (*minitr)/dxy;
}


void WindowStats(std::ofstream & outstream, Sequence::SimData & allpol, const unsigned samplesize, const unsigned windowsize, const unsigned stepsize, const unsigned chromlength)
{
  try
  {
    // calculate window-based diversity stats for a single population:
    double windowval = windowsize / (double)chromlength;
    double stepval = stepsize / (double)chromlength;
    Sequence::PolyTableSlice<Sequence::SimData> windows(allpol.sbegin(),allpol.send(),windowval,stepval,0.,1.);
    std::cout<<"Total length: "<<chromlength<<", number of windows: "<<windows.size()<<std::endl;
    for(unsigned i = 0; i < windows.size(); ++i)
    {
      Sequence::SimData window(windows[i]);
      unsigned start = i * stepsize;
      unsigned end = (start + windowsize > chromlength) ? chromlength : start + windowsize;
      double windowlength = end - start;
      unsigned nsnps = window.numsites();
      if (nsnps>0)
      {
        Sequence::PolySIM analyze(&window);
        outstream <<start<<'\t'<<end<<'\t'<<nsnps<<'\t'
                  <<analyze.ThetaPi() / windowlength<<'\t'<<analyze.ThetaW() / windowlength<<'\t'
                  <<analyze.ThetaH() / windowlength<<'\t'<<analyze.ThetaL() / windowlength<<'\t'
                  <<analyze.TajimasD()<<'\t'<<analyze.Hprime()<<std::endl;
      }
      else outstream <<start<<'\t'<<end<<'\t'<<0<<'\t'<<0<<'\t'<<0<<'\t'<<0<<'\t'<<0<<'\t'<<0<<'\t'<<0<<std::endl;
    }
  } catch (Sequence::SeqException & error){ std::cout<<error<<std::endl; }
}

void WindowStats(std::ofstream & outstream, Sequence::SimData & allpol, const unsigned samplesize1, const unsigned samplesize2, const unsigned windowsize, const unsigned stepsize, const unsigned chromlength)
{
  try
  {
    unsigned config[2] = {samplesize1,samplesize2};

    // calculate window-based diversity and differentation stats for two populations:
    double windowval = windowsize / (double)chromlength;
    double stepval = stepsize / (double)chromlength;
    Sequence::PolyTableSlice<Sequence::SimData> windows(allpol.sbegin(),allpol.send(),windowval,stepval,0,1.);
    std::cout<<"Total length: "<<chromlength<<", number of windows: "<<windows.size()<<std::endl;
    for(unsigned i = 0; i < windows.size(); ++i)
    {
      Sequence::SimData window(windows[i]);
      unsigned start = i * stepsize;
      unsigned end = (start + windowsize > chromlength) ? chromlength : start + windowsize;
      double windowlength = end - start;
      unsigned nsnps = window.numsites();
      if (nsnps>0)
      {
        Sequence::SimData wpop1(window.GetPositions(),std::vector<std::string>(window.cbegin(),window.cbegin()+samplesize1) );
        Sequence::SimData wpop2(window.GetPositions(),std::vector<std::string>(window.cbegin()+samplesize1, window.cend()) );
        Sequence::PolySIM analyze1(&wpop1);
        Sequence::PolySIM analyze2(&wpop2);
        Sequence::FST fst_neut(&window,2,config);
        outstream <<start<<'\t'<<end<<'\t'<<nsnps<<'\t'<<analyze1.ThetaPi() / windowlength<<'\t'<<analyze2.ThetaPi() / windowlength<<'\t'
                  <<analyze1.TajimasD()<<'\t'<<analyze2.TajimasD()<<'\t'<<analyze1.Hprime()<<'\t'<<analyze2.Hprime()<<'\t'
                  <<fst_neut.HSM()<<'\t'<<fst_neut.piS() / windowlength<<'\t'<<fst_neut.piB() / windowlength<<'\t'
                  <<fst_neut.piT() / windowlength<<'\t'<<calculateGmin(wpop1,wpop2)<<std::endl;
      }
      else outstream <<start<<'\t'<<end<<'\t'<<0<<'\t'<<0<<'\t'<<0<<'\t'<<0<<'\t'<<0<<'\t'<<0<<'\t'<<0<<'\t'<<"NA"
                       <<'\t'<<0<<'\t'<<0<<'\t'<<0<<"NA"<<'\t'<<std::endl;
    }

    // calculate pop subdivision stats:
    if(!allpol.empty())
    {
      Sequence::FST fst_neut(&allpol,2,config);
      std::pair< std::set<double>,std::set<double> > pneut = fst_neut.Private(0,1);
      std::cout << fst_neut.HSM() << '\t'
                << fst_neut.shared(0,1).size() << '\t'
                << pneut.first.size() << '\t'
                << pneut.second.size() << std::endl;
    }
    else std::cout << "NA\t0\t0\t0" << std::endl;
  } catch (Sequence::SeqException & error){ std::cout<<error<<std::endl; }
}

#endif

