/*
  Single population with neutral and selected mutations.
  g++ -DNDEBUG -I. -I.. -O3 -std=c++11 -o bgs_bywindow background_selection_bywindow.cc -lsequence -lgsl -lgslcblas -lz -ltcmalloc
 */

#include <config.h>
#include <fwdpp/diploid.hh>

#include <numeric>
#include <functional>
#include <cassert>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <time.h>

#include <Sequence/SimData.hpp>
#include <Sequence/PolySIM.hpp>
#include <Sequence/SeqExceptions.hpp>

#define SINGLEPOP_SIM

#include <fwdpp/sugar/sampling.hpp>
#include <fwdpp/sugar/infsites.hpp>
using mtype = KTfwd::popgenmut;
#include <common_ind.hpp>
#include <WindowStats.hpp>
using poptype = KTfwd::singlepop<mtype>;
using gcont = poptype::gcont_t;
using mcont = poptype::mcont_t;
using mcount = poptype::mcount_t;
using gtype = poptype::gamete_t;
using lookup_table_type = poptype::lookup_table_t;
using queue = std::queue<std::size_t>;



int main(int argc, char ** argv)
{
  std::cout<<"Written using fwdpp version "<<PACKAGE_VERSION<<std::endl;
  if (argc != 13)
  {
    std::cerr << "Too few arguments.\n"
              << "Usage: N mu_neutral mu_selected r s h ngens samplesize chromlength outfile_prefix nreps seed\n";
    exit(10);
  }
  int argument=1;
  const unsigned N = atoi(argv[argument++]);
  const double mu_neutral = atof(argv[argument++]);
  const double mu_selected = atof(argv[argument++]);
  const double littler = atof(argv[argument++]);
  const double s = atof(argv[argument++]);
  const double h = atof(argv[argument++]);
  const unsigned ngens = atoi(argv[argument++]); //generations to evolve
  const unsigned samplesize = atoi(argv[argument++]);
  const unsigned chromlength = atoi(argv[argument++]);
  const char * outprefix = argv[argument++];
  unsigned nreps = atoi(argv[argument++]);
  const unsigned seed = atoi(argv[argument++]);

  // Initiate random number generation system (via fwdpp/sugar/GSLrng_t.hpp)
  GSLrng r(seed);

  // Calculate size of temporary containers for population initialization
  const double theta = (mu_neutral + mu_selected) * double(4*N);
  size_t reserve_size = size_t(std::ceil( std::log(2*N)*theta+0.667*theta) );

  std::string resultfile(outprefix);
  resultfile=resultfile + "_overall.txt";
  std::ofstream resultsstream( resultfile.c_str() );

  unsigned countit=0;
  while(nreps--)
  {
    std::cout<<"Starting iteration "<<++countit<<"."<<std::endl;
    clock_t itstarttime=clock();

    // Initialize a single population
    poptype singlepop(N, reserve_size);
    unsigned generation=0; 
    unsigned interval=1000;

    std::cout<<"Starting population simulation."<<std::endl;
    clock_t simstarttime=clock();
    for( ; generation < ngens; ++generation )
    {
      if (generation % interval==0) std::cout<<"Working on generation "<<generation<<" ..."<<std::endl;
      double wbars = KTfwd::sample_diploid(r.get(),
          singlepop.gametes,
          singlepop.diploids,
          singlepop.mutations,
          singlepop.mcounts,
          N,  // N_current
          N,  // N_next
          mu_neutral + mu_selected,
          std::bind(KTfwd::infsites(),std::placeholders::_1,std::placeholders::_2,r.get(),std::ref(singlepop.mut_lookup),generation,
                    mu_neutral,mu_selected,[&r](){return gsl_rng_uniform(r.get());},[&s](){return s;},[&h](){return h;}),
          std::bind(KTfwd::poisson_xover(),r.get(),littler,0.,1.,std::placeholders::_1,std::placeholders::_2,std::placeholders::_3),
          std::bind(KTfwd::multiplicative_diploid(),std::placeholders::_1,std::placeholders::_2,std::placeholders::_3,1.),
          singlepop.neutral,
          singlepop.selected);
      KTfwd::update_mutations(singlepop.mutations,singlepop.fixations,singlepop.fixation_times,singlepop.mut_lookup,
                              singlepop.mcounts,generation,2*N);
    }



#ifndef NDEBUG
    if ( !check_sum(singlepop.gametes, 2*N) ) std::cout<<"Error: Wrong sum of gamete counts in generation "<<generation<<std::endl;
    unsigned gameteindex=0, sumcounts=0;
    std::for_each(singlepop.gametes.begin(), singlepop.gametes.end(), [&](const gtype & gamete)
    {
      sumcounts+=gamete.n;
      std::cout<<"Gamete "<<gameteindex++<<": "<<gamete.n<<std::endl;
    } );
    std::cout<<"Sum of gamete counts: "<<sumcounts<<std::endl;
    if ( singlepop.fixations.size()!=singlepop.fixation_times.size() ) std::cout<<"Error: Size of fixation and fixation times vector do not match in generation "<<generation<<std::endl;
    auto fixtime_ptr = singlepop.fixation_times.begin();
    for(auto i=singlepop.fixations.begin();i!=singlepop.fixations.end(), fixtime_ptr!=singlepop.fixation_times.end();++i,++fixtime_ptr)
    {
      std::cout<<"Position: "<<i->pos<<", generation: "<<i->g<<", time of fixation: "<<*fixtime_ptr<<std::endl;
    };
#endif

    clock_t simendtime=clock();
    printf ("Finished simulating population. Time required: %ld clicks (%f seconds).\n",
             simendtime-simstarttime,((float)(simendtime-simstarttime))/CLOCKS_PER_SEC);

    // Sample lineages, remove fixed variants in the sample
    KTfwd::sep_sample_t spop = KTfwd::sample_separate(r.get(),singlepop,samplesize,true);
    Sequence::SimData neutral(spop.first.begin(),spop.first.end());
    Sequence::SimData selected(spop.second.begin(),spop.second.end());

    // Calculate pop stats
    resultsstream<<N<<'\t'<<mu_neutral<<'\t'<<mu_selected<<'\t'<<littler<<'\t'<<s<<'\t'<<h<<'\t'<<ngens<<'\t';
    try
    {
      if(!neutral.empty())
      {
 /*     std::string msfile(outprefix);
        msfile=msfile + "_iter" + std::to_string(countit) + "_ms.txt";
        std::ofstream msstream( msfile.c_str() );
        msstream<<neutral;
        msstream.close();
*/
        Sequence::PolySIM overall(&neutral);
        resultsstream<<overall.NumPoly()<<'\t'<<overall.ThetaPi() / chromlength<<'\t'<<overall.ThetaW() / chromlength<<'\t'
                    <<overall.ThetaH() / chromlength<<'\t'<<overall.ThetaL() / chromlength<<'\t'
                    <<overall.TajimasD()<<'\t'<<overall.Hprime()<<'\t';

        std::string outfile(outprefix);
        outfile=outfile + "_iter" + std::to_string(countit) + ".bed";
        std::ofstream outstream( outfile.c_str() );
        WindowStats(outstream,neutral,samplesize,200000,200000,chromlength);
        outstream.close();
      }
      else
      {
        resultsstream<<0<<'\t'<<0<<'\t'<<0<<'\t'<<0<<'\t'<<0<<'\t'<<0<<'\t'<<0<<'\t';
      }
      if(!selected.empty())
      {
        Sequence::PolySIM overall(&selected);
        resultsstream<<overall.NumPoly()<<'\t'<<overall.ThetaPi() / chromlength<<'\t'<<overall.ThetaW() / chromlength<<'\t'
                     <<overall.ThetaH() / chromlength<<'\t'<<overall.ThetaL() / chromlength<<'\t'
                     <<overall.TajimasD()<<'\t'<<overall.Hprime()<<std::endl;
      }
      else
      {
        resultsstream<<0<<'\t'<<0<<'\t'<<0<<'\t'<<0<<'\t'<<0<<'\t'<<0<<'\t'<<0<<std::endl;
      }
    } catch (Sequence::SeqException & error){ std::cout<<error<<std::endl; }

    clock_t itendtime=clock();
    printf ("Finished iteration %u. Time required: %ld clicks (%f seconds).\n",
             countit, itendtime-itstarttime,((float)(itendtime-itstarttime))/CLOCKS_PER_SEC);
  }

resultsstream.close();
}

