/*
  Population split, period of isolation, followed by secondary contact.
  g++ -DNDEBUG -I. -I.. -O3 -std=c++11 -o scenarios scenarios.cc -lsequence -lgsl -lgslcblas -lz -ltcmalloc
 */

#include <config.h>
#include <fwdpp/diploid.hh>

#include <numeric>
#include <functional>
#include <cassert>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <utility>
#include <zlib.h>
#include <time.h>
#include <exception>

#include <Sequence/SimData.hpp>
#include <Sequence/SimDataIO.hpp> //for writing & reading SimData objects in binary format
#include <Sequence/PolySIM.hpp>
#include <Sequence/FST.hpp>
#include <Sequence/PolyTableSlice.hpp>
#include <Sequence/SeqExceptions.hpp>

#define METAPOP_SIM

#include <fwdpp/sugar/sampling.hpp>
#include <fwdpp/sugar/generalmut.hpp>
using mtype = KTfwd::generalmut<3>;
#include <IOFunctions.hpp>
#include <common_ind.hpp>
#include <Recmap.hpp>
#include <Exonmap.hpp>
#include <WindowStats.hpp>
using poptype = KTfwd::metapop<mtype>;
using gcont = poptype::gcont_t;
using mcont = poptype::mcont_t;
using mcount = poptype::mcount_t;
using gtype = poptype::gamete_t;
using diploid_bucket = poptype::dipvector_t;
using diploid_bucket_vec = poptype::vdipvector_t;
using lookup_table_type = poptype::lookup_table_t;
using queue = std::queue<std::size_t>;
#include <Mutationtracker.hpp>


std::size_t mutations_by_exonmap(queue & recycling_bin, mcont & mutations, gsl_rng * r, const KTfwd::uint_t & generation,
                   std::function<double(double)> exonmap,
                   const double & prop_weaksel, const double & prop_strongsel,
                   const double & sweak, const double & sanc, const double & s1, const double & s2,
                   const double & h,
                   bool rndpol,
                   lookup_table_type & lookup)
{
  double pos = gsl_rng_uniform(r);
  while( lookup.find(pos) != lookup.end() )
  {
    pos = gsl_rng_uniform(r);
  }
  lookup.insert(pos);
  assert(std::find_if(mutations.begin(),mutations.end(),std::bind(KTfwd::mutation_at_pos(),std::placeholders::_1,pos)) == mutations.end());

  std::array<double,3> sarray{0.,0.,0.};
  std::array<double,3> harray{0.,0.,0.};

  double exondensity = exonmap(pos);
  double prop_sel = prop_weaksel + prop_strongsel;
  if (gsl_rng_uniform(r) < exondensity*prop_sel)
  {
    harray = {h,h,h};
    if (gsl_rng_uniform(r) < prop_weaksel/prop_sel) sarray = {sweak,sweak,sweak};
    else
    {
      if (rndpol && gsl_rng_uniform(r) >= 0.5) sarray = {sanc,s2,s1};
      else sarray = {sanc,s1,s2};
    }
  }
  return KTfwd::fwdpp_internal::recycle_mutation_helper(recycling_bin,mutations,sarray,harray,pos,generation);
}


size_t migpop(const size_t & source_pop, gsl_rng * r, const double & mig_prob)
{
  if( gsl_rng_uniform(r) < mig_prob )
  {
    return ! source_pop;
  }
  return source_pop;
}


//fitness model for specific population:
struct multiplicative_diploid_pop
{
  multiplicative_diploid_pop(const unsigned __pop = 0) : popindex(__pop){};
  const unsigned popindex;
  using result_type = double;
  inline double operator()(const gtype & g1, const gtype & g2, const mcont & mutations, const double scaling = 1.) const noexcept
  {
    using __mtype =  typename mcont::value_type;
    return std::max(0., KTfwd::site_dependent_fitness()(g1,g2,mutations,
          [&](double & fitness, const __mtype & mut){ fitness *= (1. + scaling*mut.s[popindex]); },
          [&](double & fitness, const __mtype & mut){ fitness *= (1. + mut.h[popindex]*mut.s[popindex]); },
          1.));
  }
};



int main(int argc, char ** argv) try
{
  std::cout<<"Written using fwdpp version "<<PACKAGE_VERSION<<std::endl;
  if (argc != 25)
  {
    std::cerr << "Too few arguments.\n"
              << "Usage: exonmap_filename recmap_filename "
              << "scaling_factor N theta "
              << "prop_weakly_selected 2Ns_weak "
              << "prop_strongly_selected 2Ns_strong_ancestral 2Ns_strong_pop1 2Ns_strong_pop2 h random_polarize_strongmuts "
              << "ngens_before_split ngens_after_split ngens_secondary_contact 4Nm "
              << "samplesize nreps ancestralpop_file outfile_prefix seed\n";
    exit(1);
  }
  int argument=1;
  const char * exonmapfile=argv[argument++];
  const char * recmapfile=argv[argument++];
  const double scaling=atof(argv[argument++]);
  const unsigned N = atoi(argv[argument++]);
  const double theta = atof(argv[argument++]);
  const unsigned M = atoi(argv[argument++]);
  const double prop_weaksel = atof(argv[argument++]);
  const double Ns_weaksel = atof(argv[argument++]);
  const double prop_strongsel = atof(argv[argument++]);
  const double Ns_ancestral = atof(argv[argument++]);
  const double Ns_pop1 = atof(argv[argument++]);
  const double Ns_pop2 = atof(argv[argument++]);
  const double h = atof(argv[argument++]);
  const bool rndpolarization = (1 == atoi(argv[argument++]) );
  const unsigned ngens1 = atoi(argv[argument++]); //generations to evolve before split
  const unsigned ngens2 = atoi(argv[argument++]); //generations to evolve after split
  const unsigned ngens3 = atoi(argv[argument++]); //generations to evolve after secondary contact
  const unsigned samplesize = atoi(argv[argument++]);
  int nreps = atoi(argv[argument++]);
  const char * ancpopfile = argv[argument++];
  const char * outprefix = argv[argument++];
  const unsigned windowsize = atoi(argv[argument++]);
  const unsigned stepsize = atoi(argv[argument++]);
  const unsigned seed = atoi(argv[argument++]);

  const double mu = theta/double(4*N);
  size_t reserve_size = size_t(std::ceil( std::log(2*N)*theta+0.667*theta) );
  const double s_weaksel = Ns_weaksel/double(2*N);
  const double s_ancestral = Ns_ancestral/double(2*N);
  const double s_pop1 = Ns_pop1/double(2*N);
  const double s_pop2 = Ns_pop2/double(2*N);
  const unsigned metapopN []={N, N};
  const double littlem = M/double(4*N);

  // Initiate random number generation system (via fwdpp/sugar/GSLrng_t.hpp)
  GSLrng r(seed);

  // Define output file names
  std::string saveprefix(ancpopfile);
  saveprefix = saveprefix.substr(0,saveprefix.find_last_of(".") );

  // Create fitness function for ancestral population
  std::function<double (const gtype &,const gtype &,const mcont &)> vbf_anc = std::bind(multiplicative_diploid_pop(0), std::placeholders::_1, std::placeholders::_2, std::placeholders::_3, 1.);

  // Create a vector of fitness functions for each population
  std::vector<std::function<double (const gtype &,const gtype &,const mcont &)> > vbf_metapop;
  vbf_metapop.push_back(std::bind(multiplicative_diploid_pop(1),std::placeholders::_1,std::placeholders::_2,std::placeholders::_3,1.));
  vbf_metapop.push_back(std::bind(multiplicative_diploid_pop(2),std::placeholders::_1,std::placeholders::_2,std::placeholders::_3,1.));

  // Create recombination map object
  RecombinationMap recmap(r.get());
  recmap.ReadfromFile(recmapfile);
  double littler=recmap.getlittler() * scaling;
  std::cout<<"Mean recombination rate over locus per diploid: "<<littler<<"."<<std::endl;

  // Define recombination policy
  auto recmapfun = [&](const gtype & , const gtype & , const mcont &)
  {
    unsigned nbreaks = (littler > 0) ? gsl_ran_poisson(r.get(),littler) : 0u;
    std::vector<double> pos;
    if(nbreaks)
    {
      pos.reserve(nbreaks+1);
      for(unsigned i = 0 ; i < nbreaks ; ++i)
      {
        size_t cat = gsl_ran_discrete(r.get(), recmap.preproc);
        pos.emplace_back( gsl_ran_flat(r.get(), recmap.positions[cat].first, recmap.positions[cat].second) );
      }
      std::sort(pos.begin(),pos.end());
      pos.emplace_back(std::numeric_limits<double>::max());
    }
    return pos;
  };

  // Create exon map object
  ExonMap exonmap(recmap.chromlength);
  std::function<double(double)> exonmapfun;
  if (std::string(exonmapfile) == "None") exonmapfun = [](double pos){ return 1.0; };
  else
  {
    exonmap.ReadDensities(exonmapfile);
    exonmapfun = [&](double pos)
    {
      auto exonitr = upper_bound(exonmap.exondensities.begin(), exonmap.exondensities.end(), pos,
                               [](const double val, std::pair<double, double> & interval){ return val < interval.first; } );
      return ( exonitr == exonmap.exondensities.end() ) ? exonmap.exondensities.rbegin()->second : exonitr->second;
    };
  };

  // Initialize a single population in the metapop
  poptype metapop({N}, reserve_size);
  unsigned intgeneration = 0; 
  unsigned interval = 1000;

  gzFile gzf = gzopen(ancpopfile,"rb");
  if (gzclose(gzf) == Z_OK) std::cout<<"Found file with ancestral population. Skipping simulation of ancestral population."<<std::endl;
  else
  {
    std::cout<<"No valid binary gzfile with ancestral population found. Starting simulating ancestral population."<<std::endl;
    clock_t ancsimstarttime=clock();
    for( ; intgeneration < ngens1; ++intgeneration )
    {
      if (intgeneration % interval==0) std::cout<<"Working on generation "<<intgeneration<<" ..."<<std::endl;
      double wbars = KTfwd::sample_diploid(r.get(),
          metapop.gametes,
          metapop.diploids[0],
          metapop.mutations,
          metapop.mcounts,
          N,  // N_current
          N,  // N_next
          mu,
          std::bind(mutations_by_exonmap,std::placeholders::_1,std::placeholders::_2,
                    r.get(),intgeneration,exonmapfun,prop_weaksel,prop_strongsel,s_weaksel,s_ancestral,s_pop1,s_pop2,h,
                    rndpolarization,std::ref(metapop.mut_lookup) ),
          recmapfun,
          vbf_anc,
          metapop.neutral,
          metapop.selected);
      KTfwd::update_mutations(metapop.mutations,metapop.fixations,metapop.fixation_times,metapop.mut_lookup,metapop.mcounts,intgeneration,2*N);
    }

#ifndef NDEBUG
    if ( !check_sum(metapop.gametes, 2*N) ) std::cout<<"Error: Wrong sum of gamete counts in generation "<<intgeneration<<std::endl;
    unsigned gameteindex=0, sumcounts=0;
    std::for_each(metapop.gametes.begin(), metapop.gametes.end(), [&](const gtype & gamete)
    {
      sumcounts+=gamete.n;
      std::cout<<"Gamete "<<gameteindex++<<": "<<gamete.n<<std::endl;
    } );
    std::cout<<"Sum of gamete counts: "<<sumcounts<<std::endl;
    if ( metapop.fixations.size()!=metapop.fixation_times.size() ) std::cout<<"Error: Size of fixation and fixation times vector do not match in generation "<<intgeneration<<std::endl;
    auto fixtime_ptr = metapop.fixation_times.begin();
    for(auto i=metapop.fixations.begin();i!=metapop.fixations.end(), fixtime_ptr!=metapop.fixation_times.end();++i,++fixtime_ptr)
    {
      std::cout<<"Position: "<<i->pos<<", generation: "<<i->g<<", time of fixation: "<<*fixtime_ptr<<std::endl;
    };
#endif

    clock_t ancsimendtime=clock();
    printf ("Finished simulating ancestral population. Time required: %ld clicks (%f seconds).\n",
                ancsimendtime-ancsimstarttime,((float)(ancsimendtime-ancsimstarttime))/CLOCKS_PER_SEC);

    // Write ancestral population to binary gzfile.
    std::cout<<"Writting ancestral population to gzfile..."<<std::endl;
    clock_t writestarttime=clock();
    gzFile gzf = gzopen(ancpopfile,"wb");
    if (gzf == Z_NULL){ std::cerr << "Could not open file with ancestral population "<<ancpopfile<<"!"<<std::endl; exit(1); }
    KTfwd::gzserialize()(gzf,metapop,std::bind(KTfwd::mutation_writer(),std::placeholders::_1,std::placeholders::_2));
    gzclose(gzf);
    clock_t writeendtime=clock();
    printf ("Finished writing ancestral population to file. Time required: %ld clicks (%f seconds).\n",
                writeendtime-writestarttime,((float)(writeendtime-writestarttime))/CLOCKS_PER_SEC);

    // Obtain some population statistics.
    unsigned long ngametes = metapop.gametes.size();
    unsigned long ndiploids = metapop.diploids[0].size();
    unsigned long nmutations = metapop.mutations.size();
    unsigned long nneutmut = 0, nselmut = 0;
    std::for_each(metapop.mutations.begin(),metapop.mutations.end(),[&](const mtype & mut)
    {
      (mut.neutral) ? ++nneutmut : ++nselmut;
    } );
    std::cout << "Finished processing ancestral population. Number of gametes: "<<ngametes
              << ", number of mutations: "<<nmutations<<" (neutral: "<<nneutmut<<", selected: "<<nselmut<<")."<<std::endl;
  }

  // Now start iterations.
  unsigned countit=0;
  while(nreps--)
  {
    unsigned migindex=countit / nreps;
    std::cout<<"Starting iteration "<<++countit<<"."<<std::endl;
    clock_t itstarttime=clock();

    // Restore ancestral population from binary gzfile.
    clock_t readstarttime=clock();
    gzFile gzf = gzopen(ancpopfile,"rb");
    if (gzf == Z_NULL){ std::cerr << "Could not open file with ancestral population "<<ancpopfile<<"!"<<std::endl; exit(1); }
    KTfwd::gzdeserialize()(gzf,metapop,std::bind(KTfwd::mutation_reader<mtype>(),std::placeholders::_1));
    gzclose(gzf);
    clock_t readendtime=clock();
    printf ("Finished reading ancestral population from file. Time required: %ld clicks (%f seconds).\n",
              readendtime-readstarttime,((float)(readendtime-readstarttime))/CLOCKS_PER_SEC);

    // Make exact copy of first population.
    metapop.diploids.push_back(metapop.diploids[0]);

    // Population after split without migration.
    std::cout<<"Starting simulations after population split."<<std::endl;
    clock_t itsimstarttime = clock();
    unsigned generation = ngens1;

    for( ; generation < ngens1 + ngens2 ; ++generation )
    {
      if (generation % interval==0) std::cout<<"Working on generation "<<generation<<" ..."<<std::endl;
      std::vector<double> wbars = KTfwd::sample_diploid(r.get(),
          metapop.gametes,
          metapop.diploids,
          metapop.mutations,
          metapop.mcounts,
          metapopN, metapopN,
          mu,
          std::bind(mutations_by_exonmap,std::placeholders::_1,std::placeholders::_2,r.get(),generation,
                    exonmapfun,prop_weaksel,prop_strongsel,s_weaksel,s_ancestral,s_pop1,s_pop2,h,
                    rndpolarization,std::ref(metapop.mut_lookup) ),
          recmapfun,
          vbf_metapop,
          std::bind(migpop,std::placeholders::_1,r.get(),0),
          metapop.neutral,
          metapop.selected,
          nullptr);
      KTfwd::update_mutations(metapop.mutations,metapop.fixations,metapop.fixation_times,metapop.mut_lookup,metapop.mcounts,generation,4*N);
    }

#ifndef NDEBUG
    if ( !check_sum(metapop.gametes, 4*N) ) std::cout<<"Error: Wrong sum of gamete counts in generation "<<generation<<std::endl;
#endif

    clock_t itsimendtime = clock();
    printf ("Finished simulating isolation phase. Time required: %ld clicks (%f seconds).\n",
                    itsimendtime-itsimstarttime,((float)(itsimendtime-itsimstarttime))/CLOCKS_PER_SEC);

    // Sample lineages and output ms format. Don't use sugar sampling function to avoid adding fixations.
    KTfwd::sample_t spop1 = ms_sample(r.get(),metapop.mutations,metapop.gametes,metapop.diploids[0],samplesize,false);
    KTfwd::sample_t spop2 = ms_sample(r.get(),metapop.mutations,metapop.gametes,metapop.diploids[1],samplesize,false);
    Sequence::SimData allpol = merge( spop1,spop2,samplesize );

    std::string msfile1(outprefix);
    msfile1 = msfile1 + "_phase1_" + std::to_string(countit) + ".txt";
    std::ofstream outstream( msfile1.c_str() );
    outstream << allpol;
    outstream.close();

    std::string outfile1(outprefix);
    outfile1 = outfile1 + "_phase1_" + std::to_string(countit) + ".bed";
    outstream.open( outfile1.c_str() );
    WindowStats(outstream,allpol,samplesize,samplesize,windowsize,stepsize,recmap.chromlength);
    outstream.close();
    clock_t printendtime=clock();
    printf ("Finished calculating summary statistics. Time required: %ld clicks (%f seconds).\n",
                 printendtime-itsimendtime,((float)(printendtime-itsimendtime))/CLOCKS_PER_SEC);

    // Population after split with secondary contact.
    std::cout<<"Starting simulations of migration phase with m = "<<littlem<<"."<<std::endl;
    itsimstarttime=clock();

    for( ; generation < ngens1 + ngens2 + ngens3 ; ++generation )
    {
      if (generation % interval == 0) std::cout<<"Working on generation "<<generation<<" ..."<<std::endl;
      std::vector<double> wbars = KTfwd::sample_diploid(r.get(),
          metapop.gametes,
          metapop.diploids,
          metapop.mutations,
          metapop.mcounts,
          metapopN, metapopN,
          mu,
          std::bind(mutations_by_exonmap,std::placeholders::_1,std::placeholders::_2,r.get(),generation,
                    exonmapfun,prop_weaksel,prop_strongsel,s_weaksel,s_ancestral,s_pop1,s_pop2,h,
                    rndpolarization,std::ref(metapop.mut_lookup) ),
          recmapfun,
          vbf_metapop,
          std::bind(migpop,std::placeholders::_1,r.get(),littlem),
          metapop.neutral,
          metapop.selected,
          nullptr);
      KTfwd::update_mutations(metapop.mutations,metapop.fixations,metapop.fixation_times,metapop.mut_lookup,metapop.mcounts,generation,4*N);
    }

#ifndef NDEBUG
    if ( !check_sum(metapop.gametes, 4*N) ) std::cout<<"Error: Wrong sum of gamete counts in generation "<<generation<<std::endl;
#endif

    itsimendtime=clock();
    printf ("Finished simulating migration phase. Time required: %ld clicks (%f seconds).\n", 
                  itsimendtime-itsimstarttime,((float)(itsimendtime-itsimstarttime))/CLOCKS_PER_SEC);


    // Sample lineages and output ms format.
    spop1 = ms_sample(r.get(),metapop.mutations,metapop.gametes,metapop.diploids[0],samplesize,false);
    spop2 = ms_sample(r.get(),metapop.mutations,metapop.gametes,metapop.diploids[1],samplesize,false);
    allpol = merge( spop1,spop2,samplesize );

    std::string msfile2(outprefix);
    msfile2 = msfile2 + "_phase2_" + std::to_string(countit) + ".txt";
    outstream.open( msfile2.c_str() );
    outstream << allpol;
    outstream.close();

    std::string outfile2(outprefix);
    outfile2 = outfile2 + "_phase2_" + std::to_string(countit) + ".bed";
    outstream.open( outfile2.c_str() );
    WindowStats(outstream,allpol,samplesize,samplesize,windowsize,stepsize,recmap.chromlength);
    outstream.close();
    printendtime = clock();
    printf ("Finished calculating summary statistics. Time required: %ld clicks (%f seconds).\n",
             printendtime-itsimendtime,((float)(printendtime-itsimendtime))/CLOCKS_PER_SEC);

    clock_t itendtime=clock();
    printf ("Finished iteration %u. Time required: %ld clicks (%f seconds).\n",
             countit, itendtime-itstarttime,((float)(itendtime-itstarttime))/CLOCKS_PER_SEC);
  }

recmap.freepreproc();
} catch (std::exception & error){ std::cout<< error.what() <<std::endl; }


